/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.security.valve;


//core java classes
import java.io.IOException;
import java.security.Principal;
import java.net.URLEncoder;
import java.util.Enumeration;


//servlet classes
import javax.servlet.ServletException;


//core catalina classes for tomcat specific functions
import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.catalina.valves.ValveBase;
import org.apache.log4j.Logger;

//JBoss-SSO Framework classes
//jboss-sso-saml library
import org.jboss.security.saml.SingleSignOn;
import org.jboss.security.sso.util.SSOUtil;



/**
 * The SSOAutoLogout monitors an SSO session and performs automatic logout when user supposedly logs out of a federated site
 * This helps with federated logout across all the possibly federated sessions
 * 
 * <br/><br/>
 * This valve is to be used for non-JAAS based authentication mechanisms
 * 
 * @author Sohil Shah - sohil.shah@jboss.com
 */
public class PlainSSOAutoLogout extends ValveBase
{
    private static final Logger log = Logger.getLogger(PlainSSOAutoLogout.class);
    
    /**
     * properties
     */
    private String logoutURL = "";
    
    /**
     * 
     * @return
     */
    public String getLogoutURL()
    {
        return this.logoutURL;
    }
    
    /**
     * 
     * @param logoutLink
     */
    public void setLogoutURL(String logoutURL)
    {
        this.logoutURL = logoutURL;
    }
    
        
    /**
     * 
     */
    public void invoke(Request request, Response response)
    throws IOException, ServletException 
    {
        try
        { 
            SSOSession ssoSession = SSOSession.getSSOSession(request.getSession());            
            boolean ssoCookieFound = false;  
            String ssoToken = null;
            
            //find the SSOToken cookie
            ssoToken = SSOUtil.getSSOToken(request,SingleSignOn.SSO_TOKEN);
            if(ssoToken!=null && ssoToken.trim().length()>0)
            {
                ssoCookieFound = true;
            }
            
                        
            //start monitoring the SSO session on this site
            if(ssoCookieFound)
            {
                Principal principal = ssoSession.getPrincipal();                
                if(principal!=null && principal.getName()!=null)
                {
                    String ssoUserName = SSOUtil.getUsername(ssoToken);
                    String principalName = principal.getName();
                    
                    //start monitoring this session, if the current session matches the SSOToken's principal,
                    //making it an SSO session
                    if(ssoUserName!=null && principalName!=null)
                    {
                        if(ssoUserName.equals(principalName))
                        {
                            ssoSession.setMonitor(new Boolean(true));
                        }
                        else //ssoToken is out-of-sync with the actual active ssoSession on this site
                        {
                            //this happens if the user logged in as another user account on a federated site
                            //this makes the SSOToken that this is site is operating under out-of-sync with the
                            //actual SSOToken in circulation.                            
                            //this is a rare usecase,but is possible if the same user has more than one user accounts 
                            //on the group of federated sites
                            String requestURL = this.getFullRequestURL(request);
                            if(
                                    ssoCookieFound  &&  //this check if for clarity of logic, you will not get here unless this condition is true
                                    requestURL.indexOf(this.logoutURL)==-1 && //make sure this is not a request to "actually logout"..otherwise this leads to infinite redirects
                                    ssoSession.getMonitor()!=null //session is currently being monitored hence it exists
                            )
                            {
                                //step1: issue a request to logout from the site
                                if(this.signout(request,response,ssoSession))
                                {
                                    return;
                                }
                            }
                        }
                    }                    
                }
            }
            
                        
            //check if the SSO Session needs to be destroyed because the user logged out of one of the federated sites
            //this is if the SSOCookie is not available and the SSOSession Monitor is still active 
            String requestURL = this.getFullRequestURL(request);
            if(
                    !ssoCookieFound &&
                    requestURL.indexOf(this.logoutURL)==-1 && //make sure this is not a request to "actually logout"..otherwise this leads to infinite redirects
                    ssoSession.getMonitor()!=null //session is currently being monitored
                    
            )
            {
                //this means the SSOCookie is destroyed by logout(on a federated partner site), but the SSOSession is still active here
                //on this current site. So perform an automatic logout on this site as well                                
                if(this.signout(request,response,ssoSession))
                {
                    return;
                }
            }
            
            
            this.getNext().invoke(request,response);
            
        }
        catch(Exception ssoe)
        {
            log.error(this,ssoe);
            throw new ServletException(ssoe);
        }
    }  
    
    /**
     * This methods returns true if the signout was successfully performed
     * 
     * @param request
     * @param response
     * @param ssoSession
     * @param ssoToken
     * @param propagateToken
     * @return
     * @throws Exception
     */
    private boolean signout(Request request,Response response,SSOSession ssoSession) throws Exception
    {
        boolean signoutPerformed = false;
        
        //append the requested resource as a target to signout URL.
        //this way the signout operation can process it and decide on the landing page if it wants to                   
        String contextPath = ssoSession.getContext();
        String requestContext = request.getContextPath();
        if(contextPath!=null && contextPath.equals(requestContext))
        {
            String signout = this.logoutURL;
            if(this.logoutURL.indexOf('?')==-1)
            {
                signout += "?";
                //setup the requestURL as target
                signout += "target="+URLEncoder.encode(this.getRelativeRequestURL(request),"UTF-8");
            }
            else
            {
                signout += "&";
                //setup the requestURL as target
                signout += "target="+URLEncoder.encode(this.getRelativeRequestURL(request),"UTF-8");
            }                                                                    
            
            //reset the SSOSession....this when a logout operation in the application does not necessarily kill that particular active session
            ssoSession.reset();
            
                        
            //send the response
            response.getOutputStream().write(this.getLogoutMessage(signout,request).getBytes());
            response.getOutputStream().flush();
            
            signoutPerformed = true;
        }
        
        return signoutPerformed;
    }
        
    /**
     * 
     * @param request
     * @return
     */
    private String getFullRequestURL(Request request)
    {
        String fullRequestURL = "";
        
        StringBuffer buffer = request.getRequestURL();
        
        if(request.getQueryString()!=null && request.getQueryString().trim().length()>0)
        {
            buffer.append("?"+request.getQueryString());
        }
        
        fullRequestURL = buffer.toString();
        
        return fullRequestURL;
    }
    
    /**
     * 
     * @param request
     * @return
     */
    private String getRelativeRequestURL(Request request)
    {
        String relativeRequestURL = "";
        
        StringBuffer buffer = new StringBuffer(request.getRequestURI());
        
        if(request.getQueryString()!=null && request.getQueryString().trim().length()>0)
        {
            buffer.append("?"+request.getQueryString());
        }
        
        relativeRequestURL = buffer.toString();
        
        return relativeRequestURL;
    }
    
    /**
     * 
     * @param signoutURL
     * @return
     */
    private String getLogoutMessage(String signoutURL,Request request)
    {
        String logoutMessage = "";
        
        StringBuffer buffer = new StringBuffer();
        buffer.append("<html>"+"\n");
        buffer.append("<head>"+"\n");
        buffer.append("</head>"+"\n");
        buffer.append("<body onload=\"setTimeout('document.form1.submit()',1000);\">"+"\n");
        buffer.append("<form name=\"form1\" action=\""+signoutURL+"\" method=\"post\">"+"\n");
        Enumeration parameterNames = request.getParameterNames();
        while(parameterNames.hasMoreElements())
        {
        	String name = (String)parameterNames.nextElement();
        	String value = request.getParameter(name);
        	
        	//exclude some sso control control parameters from the post params
        	if(name.equals("token") || name.equals("target"))
        	{
        		continue;
        	}
        	
        	buffer.append("<input type=\"hidden\" name=\""+name+"\" value=\""+value+"\"/>"+"\n");
        }
        buffer.append("</form>"+"\n");
        buffer.append("</body>"+"\n");
        buffer.append("</html>"+"\n");
        
        
        logoutMessage = buffer.toString();
        
        return logoutMessage;
    }
}
